import requests

from ....sdk.contracts.dtos.sdk.responses.response import Response
from ....sdk.contracts.dtos.sdk.responses.worlds.get import WorldsGetResponse
from ..abstract import AbstractCommand


class WorldsGetCommand(AbstractCommand):
    async def execute(self) -> WorldsGetResponse:
        response: requests.Response = requests.get(
            url=f"http://{self.options.tld}/worlds", timeout=self.options.timeout
        )
        return Response[WorldsGetResponse].model_validate(response.json()).data



import logging
from enum import Enum

from ......sdk.contracts.dtos.tiles.address import Address
from ......server.clients.dao import DaoClient
from .....common.utils import generate_random_float
from ....types.entity import EntityType
from ..entity import Entity

logger = logging.getLogger()


class EntityFungi(Entity):
    address: Address
    daoclient: DaoClient

    class Meta(Enum):
        SPORE = 0
        MYCELIUM = 1
        FRUIT = 2
        MATURE = 3

    entity_type: EntityType = EntityType.FUNGI

    max_amount: int = 16
    mutation_rate: float = 0.01

    async def quantum(self) -> None:
        """ """
        # logger.info("fungi entity quantum")

        if EntityFungi.Meta(self.state) == EntityFungi.Meta.SPORE:
            if generate_random_float() <= self.mutation_rate:
                self.state = EntityFungi.Meta.MYCELIUM.value
                await self.daoclient.patch(address=self.address, document={"state": self.state})
        elif EntityFungi.Meta(self.state) == EntityFungi.Meta.MYCELIUM:
            if generate_random_float() <= self.mutation_rate:
                self.state = EntityFungi.Meta.FRUIT.value
                await self.daoclient.patch(address=self.address, document={"state": self.state})
        elif EntityFungi.Meta(self.state) == EntityFungi.Meta.FRUIT:
            if generate_random_float() <= self.mutation_rate:
                self.state = EntityFungi.Meta.MATURE.value
                await self.daoclient.patch(address=self.address, document={"state": self.state})
        elif EntityFungi.Meta(self.state) == EntityFungi.Meta.MATURE:
            if generate_random_float() <= self.mutation_rate:
                if self.amount < self.max_amount:
                    self.amount = self.amount + 1
                    await self.daoclient.patch(address=self.address, document={"amount": self.amount})
                self.state = EntityFungi.Meta.SPORE.value
                await self.daoclient.patch(address=self.address, document={"state": self.state})



from pydantic import BaseModel

from ...types.entity import EntityType


class AbstractEntity(BaseModel):
    id: str
    name: str | None = None
    rbac: dict = {}
    entity_type: EntityType = EntityType.UNKNOWN

    amount: float = 0
    state: int = 0



import asyncio
import logging
import uuid
from asyncio import Queue

from pydantic import BaseModel

from ....sdk.contracts.dtos.entities.entity import Entity
from ....sdk.contracts.dtos.entities.fauna.fish import EntityFish
from ....sdk.contracts.dtos.entities.flora.grass import EntityGrass
from ....sdk.contracts.dtos.entities.flora.tree import EntityTree
from ....sdk.contracts.dtos.entities.funga.fungi import EntityFungi
from ....sdk.contracts.dtos.tiles.address import Address
from ....sdk.contracts.dtos.tiles.tile import Tile
from ....sdk.contracts.types.entity import EntityType
from ....sdk.contracts.types.tile import TileType
from ...clients.dao import DaoClient

logger = logging.getLogger()


class AbstractEntityFactory(BaseModel):
    daoclient: DaoClient

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    async def producer(ids: set[str], queue: Queue):
        for local_id in ids:
            await queue.put(local_id)

    async def generate(self, address: Address) -> None:
        # get entities ids for the tile
        local_tile: Tile = await self.daoclient.get(address=address)

        # if len(local_tile.ids) > 0:
        #     msg: str = f"entity generation can occurring on a tile with pre-existing entities, {address}"
        #     logger.warning(msg)

        # Review types now
        if local_tile.tile_type == TileType.GRASS:
            new_entity: Entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.GRASS)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

        elif local_tile.tile_type == TileType.FOREST:
            new_entity: Entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.TREE)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

            new_entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.FUNGI)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

        elif local_tile.tile_type == TileType.OCEAN:
            new_entity: Entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.FISH)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

        patch: dict = {"ids": list(local_tile.ids)}
        # print(f"tile id {local_tile}, got: {patch}")
        await self.daoclient.patch(address=address, document=patch)

    ###

    async def grow_entities(self, address: Address):
        # get entities ids for the tile
        local_tile: Tile = await self.daoclient.get(address=address)

        async def entity_producer(queue: Queue):
            for entity_id in local_tile.ids:
                await queue.put(entity_id)

        async def step_one():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_entity_id: str = await queue.get()

                    # TODO: Process entity
                    address_entity: Address = Address.model_validate(
                        {**address.model_dump(), "entity_id": local_entity_id}
                    )
                    entity: Entity = await self.daoclient.get(address=address_entity)
                    # print(wrapped_entity.model_dump())

                    if entity.entity_type == EntityType.FUNGI:
                        fungi: EntityFungi = EntityFungi.model_validate(
                            {**entity.model_dump(), "address": address_entity, "daoclient": self.daoclient}
                        )
                        await fungi.quantum()
                    elif entity.entity_type == EntityType.GRASS:
                        grass: EntityGrass = EntityGrass.model_validate(
                            {**entity.model_dump(), "address": address_entity, "daoclient": self.daoclient}
                        )
                        await grass.quantum()
                    elif entity.entity_type == EntityType.FISH:
                        fish: EntityFish = EntityFish.model_validate(
                            {**entity.model_dump(), "address": address_entity, "daoclient": self.daoclient}
                        )
                        await fish.quantum()
                    elif entity.entity_type == EntityType.TREE:
                        tree: EntityTree = EntityTree.model_validate(
                            {**entity.model_dump(), "address": address_entity, "daoclient": self.daoclient}
                        )
                        await tree.quantum()

                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(entity_producer(queue), consumer(queue))

        await step_one()



import logging
import os

import click
import uvicorn
from fastapi import FastAPI

from ..common.routers import metrics
from .context import ContextManager
from .routers import world, worlds

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


# pylint: disable=too-many-arguments,no-value-for-parameter
@click.command()
@click.option("--hostname", type=click.STRING, default="0.0.0.0", help="monogdb host address to connect to")
@click.option("--port", type=click.INT, default=8000, help="monogdb host port to connect to")
@click.option("--mongodb-hostname", type=click.STRING, default="0.0.0.0", help="monogdb host address to connect to")
@click.option("--mongodb-port", type=click.INT, default=27017, help="monogdb host port to connect to")
@click.option("--mongodb-database", type=click.STRING, default="darkness", help="monogdb database name")
@click.option("--log-level", type=click.STRING, default="INFO", help="log level (INFO, DEBUG, WARNING, ERROR, FATAL)")
def main(
    hostname: str, port: int, mongodb_hostname: str, mongodb_port: int, mongodb_database: str, log_level: str
) -> None:
    logger.setLevel(logging.getLevelName(log_level))

    os.environ["DARKNESS_MONGODB_HOST"] = mongodb_hostname
    os.environ["DARKNESS_MONGODB_PORT"] = str(mongodb_port)
    os.environ["DARKNESS_MONGODB_DATABASE"] = mongodb_database
    ContextManager.deferred_init()

    logger.info("[Main] starting")
    app = FastAPI()

    logger.debug("[Main] adding metrics routes")
    app.include_router(metrics.router)

    logger.debug("[Main] adding world routes")
    app.include_router(world.router)

    logger.debug("[Main] adding worlds routes")
    app.include_router(worlds.router)

    logger.info("[Main] online")

    uvicorn.run(app=app, host=hostname, port=port, log_level=logging.getLevelName(log_level))


if __name__ == "__main__":
    main()



import asyncio
import json
import logging
import re
import uuid
from abc import abstractmethod
from asyncio import Queue

from pydantic import BaseModel

from ....sdk.common.utils import generate_random_float
from ....sdk.contracts.dtos.entities.entity import Entity
from ....sdk.contracts.dtos.tiles.address import Address
from ....sdk.contracts.dtos.tiles.chunk import Chunk
from ....sdk.contracts.dtos.tiles.tile import Tile
from ....sdk.contracts.dtos.tiles.world import World
from ....sdk.contracts.dtos.window import Window
from ....sdk.contracts.types.connection import TileConnectionType
from ....sdk.contracts.types.dao_document import DaoDocumentType
from ....sdk.contracts.types.tile import TileType
from ...clients.dao import DaoClient
from ..entity.entity import EntityFactory

logger = logging.getLogger()


# pylint: disable=too-many-statements
class AbstractChunkFactory(BaseModel):
    daoclient: DaoClient
    entity_factory: EntityFactory

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    async def producer(ids: set[str], queue: Queue):
        for local_id in ids:
            await queue.put(local_id)

    @abstractmethod
    async def create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> str:
        """ """

    async def mutate_tile(self, address: Address, mutate: float, tile_type: TileType) -> None:
        if generate_random_float() <= mutate:
            target_tile: Tile = await self.daoclient.get(address=address)
            await self.convert_tile(address=address, source=target_tile.tile_type, target=tile_type)

    async def convert_tile(
        self, address: Address, source: TileType, target: TileType, clear_entities: bool = False
    ) -> None:
        # get
        target_tile: Tile = await self.daoclient.get(address=address)
        # check
        if target_tile.tile_type == source:
            doc_patch: dict = {"tile_type": target}
            if clear_entities:
                for child_id in target_tile.ids:
                    await self.daoclient.delete(
                        address=Address.model_validate({**address.model_dump(), "entity_id": child_id})
                    )
                # update doc_patch
                doc_patch["ids"] = []

            await self.daoclient.patch(address=address, document=doc_patch)

            # repopulate entities
            await self.entity_factory.generate(address=address)

    async def adjacent_liquids(self, address: Address, depth: int) -> list[TileType]:
        return await self.adjecent_to(address=address, types=[TileType.OCEAN, TileType.WATER], depth=depth)

    async def adjacents(self, address: Address) -> list[Tile]:
        target_tile: Tile = await self.daoclient.get(address=address)
        adjecent_tile_addresses: list[Address] = []
        for _, adjecent_id in target_tile.next.items():
            adjecent_tile_addresses.append(Address.model_validate({**address.model_dump(), "tile_id": adjecent_id}))

        adjecent_tiles: list[Tile] = await self.daoclient.get_multi(
            addresses=adjecent_tile_addresses, doc_type=DaoDocumentType.TILE
        )
        return adjecent_tiles

    async def adjacent_recursive(self, address: Address, depth: int) -> list[Tile]:
        if depth < 0:
            return []

        adjacent_tiles = await self.adjacents(address=address)

        if depth > 1:
            child_tiles: list[Tile] = []
            for adj_tile in adjacent_tiles:
                current_children: list[Tile] = await self.adjacent_recursive(
                    address=Address.model_validate({**address.model_dump(), "tile_id": adj_tile.id}), depth=depth - 1
                )
                child_tiles = child_tiles + current_children

            adjacent_tiles = adjacent_tiles + child_tiles

        tile_ids: set[str] = set()
        deduped_tiles: list[Tile] = []
        for tile in adjacent_tiles:
            if tile.id not in tile_ids:
                tile_ids.add(tile.id)
                deduped_tiles.append(tile)
        return deduped_tiles

    async def adjecent_to(self, address: Address, types: list[TileType] | None, depth: int) -> list[TileType]:
        adjecent_targets: list[TileType] = []
        adjecent_tiles: list[Tile] = await self.adjacent_recursive(address=address, depth=depth)
        for adjecent_tile in adjecent_tiles:
            if adjecent_tile.tile_type in types:
                adjecent_targets.append(adjecent_tile.tile_type)
        return list(set(adjecent_targets))

    async def _grow_dirt_tile(self, address: Address) -> None:
        adjecent_liquids: list[TileType] = await self.adjacent_liquids(address=address, depth=1)

        # Grass does not grow directly next to the ocean
        if TileType.OCEAN not in adjecent_liquids:
            if TileType.WATER in adjecent_liquids:
                await self.mutate_tile(address=address, mutate=0.01, tile_type=TileType.GRASS)
            else:
                # Grass can not grow more than 3 tiles beyond a WATER source
                liquid_tiles: list[TileType] = await self.adjecent_to(address=address, types=[TileType.WATER], depth=3)
                if TileType.WATER in liquid_tiles:

                    # Grass grows from other grass
                    adjecent_flora: list[TileType] = await self.adjecent_to(
                        address=address, types=[TileType.FOREST, TileType.GRASS], depth=1
                    )
                    if TileType.FOREST in adjecent_flora:
                        await self.mutate_tile(address=address, mutate=0.005, tile_type=TileType.GRASS)

                    # elif len(adjecent_flora) > 0:
                    #     await self.mutate_tile(address=address, mutate=0.00125, tile_type=TileType.GRASS)

    async def _grow_grass_tile(self, address: Address) -> None:
        neighbors: list[TileType] = await self.adjecent_to(
            address=address,
            types=[TileType.WATER, TileType.GRASS, TileType.FOREST, TileType.OCEAN, TileType.DIRT],
            depth=1,
        )

        # no forests grow by oceans or bare dirt
        if TileType.OCEAN in neighbors or TileType.DIRT in neighbors:
            return

        if len(neighbors) > 1:
            await self.mutate_tile(address=address, mutate=0.00125, tile_type=TileType.FOREST)

    async def tile_grow(self, address: Address) -> None:
        # get
        target_tile: Tile = await self.daoclient.get(address=address)

        # dirt -> grass
        if target_tile.tile_type == TileType.DIRT:
            await self._grow_dirt_tile(address=address)

        # grass+water (no dirt/ocean) -> forest
        if target_tile.tile_type == TileType.GRASS:
            await self._grow_grass_tile(address=address)

    async def tile_senescence(self, address: Address) -> None:
        # get
        target_tile: Tile = await self.daoclient.get(address=address)

        ##
        if len(target_tile.ids) < 1:
            # then we don't have any entities, downgrade as appropriate
            if target_tile.tile_type == TileType.GRASS:
                await self.mutate_tile(address=address, mutate=1, tile_type=TileType.DIRT)
            elif target_tile.tile_type == TileType.FOREST:
                await self.mutate_tile(address=address, mutate=1, tile_type=TileType.GRASS)

    async def brackish_tile(self, address: Address) -> None:
        # Convert inner Ocean to Water Tiles

        # See if we are next to another ocean tile
        neighbors: list[TileType] = await self.adjecent_to(address=address, types=[TileType.OCEAN], depth=1)
        if len(neighbors) < 1:
            await self.convert_tile(address=address, source=TileType.OCEAN, target=TileType.WATER)

        # TODO: are we an isolated ocean body? if so then we are water
        #
        # # 1. For each next that is liquid (recusively), can we get to the edge?
        # #     # 1. Are we ocean? - if no then return False
        # #     # 2. Are we edge? - if yes then return True
        # #     # 3. then check my nexts
        # #     # 4. Return False (not ocean)

    async def erode_tile(self, address: Address) -> None:
        # get
        target_tile: Tile = await self.daoclient.get(address=address)

        # shore erosion
        if target_tile.tile_type not in [TileType.UNKNOWN, TileType.OCEAN, TileType.WATER, TileType.SHORE]:
            adjecent_liquids: list[TileType] = await self.adjacent_liquids(address=address, depth=1)
            # Apply erosion - rocks and be left by oceans, everything else becomes shore
            if TileType.OCEAN in adjecent_liquids:
                if target_tile.tile_type not in (TileType.ROCK, TileType.SHORE):
                    # await self.daoclient.patch(address=address, document={"tile_type": TileType.SHORE})
                    await self.convert_tile(address=address, source=target_tile.tile_type, target=TileType.SHORE)

    async def gen_geo_bind(self, address: Address, conn_dir: TileConnectionType, target_tile_id: str) -> None:
        # TODO: move to native mongodb to make this efficient ..
        doc: World | Chunk | Tile | Entity = await self.daoclient.get(address=address)
        doc.next[conn_dir] = target_tile_id
        await self.daoclient.patch(address=address, document=json.loads(doc.model_dump_json()))

    async def generate_ocean_block(self, address: Address, window: Window):
        tile_map: dict[str, str] = {}

        async def flat_producer(window: Window, queue: Queue):
            range_x_min: int = window.min.x - 1
            range_x_max: int = window.max.x
            range_y_min: int = window.min.x - 1
            range_y_max: int = window.max.y

            for x in range(range_x_min, range_x_max):
                for y in range(range_y_min, range_y_max):
                    local_x = x + 1
                    local_y = y + 1
                    tile_id: str = f"tile_{local_x}_{local_y}"
                    await queue.put(tile_id)

        # 1. fill a blank nXm area with ocean
        async def step_one():
            # get container
            async def consumer(queue: Queue):
                chunk: Chunk = await self.daoclient.get(address=address)

                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    tile_map[local_tile_id] = str(uuid.uuid4())
                    local_tile: Tile = Tile(id=tile_map[local_tile_id], tile_type=TileType.OCEAN)
                    address_tile: Address = Address.model_validate({**address.model_dump(), "tile_id": local_tile.id})

                    # create tile
                    await self.daoclient.post(address=address_tile, document=local_tile)
                    # msg: str = f"({local_tile_id}) brought into existence as {TileType.OCEAN}"
                    # logger.debug(msg)

                    # update
                    chunk.ids.add(tile_map[local_tile_id])
                    queue.task_done()

                # Update the chunk --
                # put -- store chunk update (tile addition)
                document = {"ids": list(chunk.ids)}
                # print(document)
                await self.daoclient.patch(address=address, document=document)

            queue = asyncio.Queue()
            await asyncio.gather(flat_producer(window, queue), consumer(queue))

        await step_one()

        # set origin tile on chunk
        await self.daoclient.patch(address=address, document={"origin": tile_map["tile_1_1"]})

        # 2. Connect everything together
        async def step_two():
            pattern: re.Pattern = re.compile("^tile_([a-zA-Z0-9-]+)_([a-zA-Z0-9-]+)$")

            async def consumer(queue):
                while not queue.empty():
                    _local_tile_id: str = await queue.get()
                    local_tile_id = tile_map[_local_tile_id]

                    match = pattern.search(_local_tile_id)
                    # tile_id: str = f"tile_{local_x}_{local_y}"
                    local_x: int = int(match.group(1))
                    local_y: int = int(match.group(2))

                    # msg: str = f"binding ({local_tile_id}) physically to peers"
                    # logger.debug(msg)

                    # Bind LEFT
                    _target_tile_id: str = f"tile_{local_x - 1}_{local_y}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.LEFT,
                        )

                    # Bind RIGHT
                    _target_tile_id: str = f"tile_{local_x + 1}_{local_y}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.RIGHT,
                        )

                    # Bind UP
                    _target_tile_id: str = f"tile_{local_x}_{local_y - 1}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.UP,
                        )

                    # Bind DOWN
                    _target_tile_id: str = f"tile_{local_x}_{local_y + 1}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.DOWN,
                        )

                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(flat_producer(window, queue), consumer(queue))

        await step_two()



import requests

from ....sdk.contracts.dtos.sdk.requests.chunk.quantum import ChunkQuantumRequest
from ..abstract import AbstractCommand


class ChunkQuantumCommand(AbstractCommand):
    async def execute(self, request: ChunkQuantumRequest) -> None:
        requests.post(
            url=f"http://{self.options.tld}/world/{request.world_id}/chunk/{request.chunk_id}",
            timeout=self.options.timeout,
            json=request.model_dump(exclude={"world_id", "chunk_id"}),
        )



import secrets

from ..contracts.dtos.tiles.address import Address
from ..contracts.errors.server.dao.unknown import DaoUnknownError
from ..contracts.types.dao_document import DaoDocumentType


def address_type(address: Address) -> DaoDocumentType:
    """
    Function to determine address type.

    Parameters
    ----------
    address: Address
        Address to review

    Returns
    -------
    address_type: DaoDocumentType
        The address type
    """

    if address.entity_id is not None:
        return DaoDocumentType.ENTITY

    if address.tile_id is not None:
        return DaoDocumentType.TILE

    if address.chunk_id is not None:
        return DaoDocumentType.CHUNK

    if address.world_id is not None:
        return DaoDocumentType.WORLD

    raise DaoUnknownError("invalid document type")


def get_document_id_from_address(address: Address, doc_type: DaoDocumentType | None = None) -> str:
    """
    Returns document id for a given address and type.

    Parameters
    ----------
    address: Address
        Address to review
    doc_type: DaoDocumentType
        The document type of the address

    Returns
    -------
    document_id: str
        The document id of the address/type.
    """

    if doc_type is None:
        doc_type = address_type(address)

    if doc_type == DaoDocumentType.ENTITY:
        return address.entity_id

    if doc_type == DaoDocumentType.TILE:
        return address.tile_id

    if doc_type == DaoDocumentType.CHUNK:
        return address.chunk_id

    if doc_type == DaoDocumentType.WORLD:
        return address.world_id

    raise DaoUnknownError("invalid document type")


def generate_random_float(min_val: float = 0.0, max_val: float = 1.0) -> float:
    """
    Generate a random 16-bit floating point number between `min_val` and `max_val`.

    Parameters
    ----------
    min_val: float
        The minimum value
    max_val: float
        The maximum value

    Returns
    -------
    random_float: float
        The random floating point number
    """

    random_int: int = secrets.randbelow(2**16)
    random_float: float = min_val + ((max_val - min_val) * (random_int / (2**16)))
    return random_float



"""" State Service Client Module """

from ..sdk.contracts.dtos.entities.entity import Entity
from ..sdk.contracts.dtos.sdk.command_options import CommandOptions
from ..sdk.contracts.dtos.sdk.requests.chunk.create import ChunkCreateRequest
from ..sdk.contracts.dtos.sdk.requests.chunk.delete import ChunkDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.chunk.get import ChunkGetRequest
from ..sdk.contracts.dtos.sdk.requests.chunk.patch import ChunkPatchRequest
from ..sdk.contracts.dtos.sdk.requests.chunk.quantum import ChunkQuantumRequest
from ..sdk.contracts.dtos.sdk.requests.entity.delete import EntityDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.entity.entity import EntityRequest
from ..sdk.contracts.dtos.sdk.requests.entity.patch import EntityPatchRequest
from ..sdk.contracts.dtos.sdk.requests.tile.delete import TileDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.tile.get import TileGetRequest
from ..sdk.contracts.dtos.sdk.requests.tile.patch import TilePatchRequest
from ..sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ..sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ..sdk.contracts.dtos.sdk.requests.world.patch import WorldPatchRequest
from ..sdk.contracts.dtos.sdk.responses.chunk.create import ChunkCreateResponse
from ..sdk.contracts.dtos.sdk.responses.chunk.get import ChunkGetResponse
from ..sdk.contracts.dtos.sdk.responses.entity.get import EntityGetResponse
from ..sdk.contracts.dtos.sdk.responses.tile.get import TileGetResponse
from ..sdk.contracts.dtos.sdk.responses.world.create import WorldCreateResponse
from ..sdk.contracts.dtos.sdk.responses.world.get import WorldGetResponse
from ..sdk.contracts.dtos.sdk.responses.worlds.get import WorldsGetResponse
from ..sdk.contracts.dtos.tiles.chunk import Chunk
from ..sdk.contracts.dtos.tiles.tile import Tile
from ..sdk.contracts.dtos.tiles.world import World
from ..sdk.contracts.types.chunk_quantum import ChunkQuantumType
from ..sdk.contracts.types.tile import TileType
from .commands.chunk.create import ChunkCreateCommand
from .commands.chunk.delete import ChunkDeleteCommand
from .commands.chunk.get import ChunkGetCommand
from .commands.chunk.patch import ChunkPatchCommand
from .commands.chunk.quantum import ChunkQuantumCommand
from .commands.entity.delete import EntityDeleteCommand
from .commands.entity.get import EntityGetCommand
from .commands.entity.patch import EntityPatchCommand
from .commands.metrics.health.get import HealthGetCommand
from .commands.tile.delete import TileDeleteCommand
from .commands.tile.get import TileGetCommand
from .commands.tile.patch import TilePatchCommand
from .commands.world.create import WorldCreateCommand
from .commands.world.delete import WorldDeleteCommand
from .commands.world.get import WorldGetCommand
from .commands.world.patch import WorldPatchCommand
from .commands.worlds.get import WorldsGetCommand


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-many-public-methods
class StateClient:
    # metrics/health
    health_get_command: HealthGetCommand

    # entity
    entity_delete_command: EntityDeleteCommand
    entity_get_command: EntityGetCommand
    entity_patch_command: EntityPatchCommand

    # tile
    tile_delete_command: TileDeleteCommand
    tile_get_command: TileGetCommand
    tile_patch_command: TilePatchCommand

    # chunk
    chunk_create_command: ChunkCreateCommand
    chunk_delete_command: ChunkDeleteCommand
    chunk_get_command: ChunkGetCommand
    chunk_patch_command: ChunkPatchCommand
    chunk_quantum_command: ChunkQuantumCommand

    # world
    world_create_command: WorldCreateCommand
    world_delete_command: WorldDeleteCommand
    world_get_command: WorldGetCommand
    world_patch_command: WorldPatchCommand

    # worlds
    worlds_get_command: WorldsGetCommand

    def __init__(self, options: CommandOptions | None = None):
        command_dict: dict = {}
        if options:
            command_dict["options"] = options

        ### metrics
        self.health_get_command = HealthGetCommand.model_validate(command_dict)

        ## entity
        self.entity_delete_command = EntityDeleteCommand.model_validate(command_dict)
        self.entity_get_command = EntityGetCommand.model_validate(command_dict)
        self.entity_patch_command = EntityPatchCommand.model_validate(command_dict)

        ## tile
        self.tile_delete_command = TileDeleteCommand.model_validate(command_dict)
        self.tile_get_command = TileGetCommand.model_validate(command_dict)
        self.tile_patch_command = TilePatchCommand.model_validate(command_dict)

        ### chunk
        self.chunk_create_command = ChunkCreateCommand.model_validate(command_dict)
        self.chunk_delete_command = ChunkDeleteCommand.model_validate(command_dict)
        self.chunk_get_command = ChunkGetCommand.model_validate(command_dict)
        self.chunk_patch_command = ChunkPatchCommand.model_validate(command_dict)
        self.chunk_quantum_command = ChunkQuantumCommand.model_validate(command_dict)

        # world
        self.world_create_command = WorldCreateCommand.model_validate(command_dict)
        self.world_delete_command = WorldDeleteCommand.model_validate(command_dict)
        self.world_get_command = WorldGetCommand.model_validate(command_dict)
        self.world_patch_command = WorldPatchCommand.model_validate(command_dict)

        # worlds
        self.worlds_get_command = WorldsGetCommand.model_validate(command_dict)

    # metrics

    async def health_get(self) -> dict:
        """
        Gets the server health.

        Returns
        -------
        health: dict
            The server health (true or false).
        """

        return await self.health_get_command.execute()

    # entity
    async def entity_delete(self, world_id: str, chunk_id: str, tile_id: str, entity_id: str, parent: bool) -> bool:
        request: EntityDeleteRequest = EntityDeleteRequest(
            world_id=world_id, chunk_id=chunk_id, tile_id=tile_id, entity_id=entity_id, parent=parent
        )
        success: bool = await self.entity_delete_command.execute(request=request)
        return success

    async def entity_get(self, world_id: str, chunk_id: str, tile_id: str, entity_id: str) -> Entity:
        request: EntityRequest = EntityRequest(
            world_id=world_id, chunk_id=chunk_id, tile_id=tile_id, entity_id=entity_id
        )
        response: EntityGetResponse = await self.entity_get_command.execute(request=request)
        return response.data

    async def entity_patch(self, world_id: str, chunk_id: str, tile_id: str, entity_id: str, partial: dict) -> bool:
        request: EntityPatchRequest = EntityPatchRequest(
            world_id=world_id, chunk_id=chunk_id, tile_id=tile_id, entity_id=entity_id, partial=partial
        )
        success: bool = await self.entity_patch_command.execute(request=request)
        return success

    # tile
    async def tile_delete(self, world_id: str, chunk_id: str, tile_id: str, parent: bool, cascade: bool) -> bool:
        request: TileDeleteRequest = TileDeleteRequest(
            world_id=world_id, chunk_id=chunk_id, tile_id=tile_id, parent=parent, cascade=cascade
        )
        success: bool = await self.tile_delete_command.execute(request=request)
        return success

    async def tile_get(self, world_id: str, chunk_id: str, tile_id: str, full: bool) -> Tile:
        request: TileGetRequest = TileGetRequest(world_id=world_id, chunk_id=chunk_id, tile_id=tile_id, full=full)
        response: TileGetResponse = await self.tile_get_command.execute(request=request)
        return response.tile

    async def tile_patch(self, world_id: str, chunk_id: str, tile_id: str, partial: dict) -> bool:
        request: TilePatchRequest = TilePatchRequest(
            world_id=world_id, chunk_id=chunk_id, tile_id=tile_id, partial=partial
        )
        success: bool = await self.tile_patch_command.execute(request=request)
        return success

    # chunk

    async def chunk_create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> str:
        request: ChunkCreateRequest = ChunkCreateRequest(
            world_id=world_id, name=name, dimensions=dimensions, biome=biome
        )
        response: ChunkCreateResponse = await self.chunk_create_command.execute(request=request)
        return response.id

    async def chunk_delete(self, world_id: str, chunk_id: str, parent: bool, cascade: bool) -> bool:
        request: ChunkDeleteRequest = ChunkDeleteRequest(
            world_id=world_id, chunk_id=chunk_id, parent=parent, cascade=cascade
        )
        success: bool = await self.chunk_delete_command.execute(request=request)
        return success

    async def chunk_get(self, world_id: str, chunk_id: str, full: bool = False) -> Chunk:
        request: ChunkGetRequest = ChunkGetRequest(world_id=world_id, chunk_id=chunk_id, full=full)
        response: ChunkGetResponse = await self.chunk_get_command.execute(request=request)
        return response.chunk

    async def chunk_patch(self, world_id: str, chunk_id: str, partial: dict) -> bool:
        request: ChunkPatchRequest = ChunkPatchRequest(world_id=world_id, chunk_id=chunk_id, partial=partial)
        success: bool = await self.chunk_patch_command.execute(request=request)
        return success

    async def chunk_quantum(self, world_id: str, chunk_id: str, scope: ChunkQuantumType) -> None:
        request: ChunkQuantumRequest = ChunkQuantumRequest(world_id=world_id, chunk_id=chunk_id, scope=scope)
        await self.chunk_quantum_command.execute(request=request)

    # world

    async def world_create(self, name: str | None) -> str:
        request: WorldCreateRequest = WorldCreateRequest(name=name)
        response: WorldCreateResponse = await self.world_create_command.execute(request)
        return response.id

    async def world_delete(self, world_id: str, cascade: bool) -> bool:
        request: WorldDeleteRequest = WorldDeleteRequest(world_id=world_id, cascade=cascade)
        success: bool = await self.world_delete_command.execute(request=request)
        return success

    async def world_get(self, world_id: str, full: bool) -> World:
        request: WorldGetRequest = WorldGetRequest(world_id=world_id, full=full)
        response: WorldGetResponse = await self.world_get_command.execute(request)
        return response.world

    async def world_patch(self, world_id: str, partial: dict) -> bool:
        request: WorldPatchRequest = WorldPatchRequest(world_id=world_id, partial=partial)
        success: bool = await self.world_patch_command.execute(request=request)
        return success

    # worlds

    async def worlds_get(self) -> list[World]:
        response: WorldsGetResponse = await self.worlds_get_command.execute()
        return response.worlds



""" Tile Get Command """

import requests

from ....sdk.contracts.dtos.sdk.requests.tile.get import TileGetRequest
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ....sdk.contracts.dtos.sdk.responses.tile.get import TileGetResponse
from ..abstract import AbstractCommand


class TileGetCommand(AbstractCommand):
    async def execute(self, request: TileGetRequest) -> TileGetResponse:
        response: requests.Response = requests.get(
            url=f"http://{self.options.tld}/world/{request.world_id}/chunk/{request.chunk_id}/tile/{request.tile_id}",
            timeout=self.options.timeout,
            params={"full": request.full},
        )
        return Response[TileGetResponse].model_validate(response.json()).data



import logging
import os
from contextlib import asynccontextmanager

import click
import uvicorn
from fastapi import FastAPI
from typing_extensions import AsyncIterator

from ...api.common.routers import metrics
from .context import ContextManager
from .routers import chrono
from .routers.chrono import world_chrono

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


# pylint: disable=unused-argument
@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator[FastAPI]:
    logger.debug("Starting lifespan")
    await world_chrono()
    yield
    logger.debug("Ending lifespan")


# pylint: disable=too-many-arguments,no-value-for-parameter
@click.command()
@click.option("--hostname", type=click.STRING, default="0.0.0.0", help="host address to bind to")
@click.option("--port", type=click.INT, default=9000, help="port to bind to")
@click.option("--log-level", type=click.STRING, default="INFO", help="log level (INFO, DEBUG, WARNING, ERROR, FATAL)")
@click.option("--darkness-hostname", type=click.STRING, default="0.0.0.0", help="state api address to connect to")
@click.option("--darkness-port", type=click.INT, default=8000, help="state api port to connect to")
@click.option("--darkness-sleep-time", type=click.FLOAT, default=1.0, help="state api call sleep time (seconds)")
@click.option("--darkness-timeout", type=click.FLOAT, default=5.0, help="state api call timeout (seconds)")
@click.option("--darkness-retries", type=click.INT, default=5, help="state api call retries (integer)")
@click.option(
    "--darkness-url", type=click.STRING, is_flag=False, default=None, help="state api url, format: '(host):(port)'"
)
def main(
    hostname: str,
    port: int,
    log_level: str,
    darkness_hostname: str,
    darkness_port: int,
    darkness_sleep_time: float,
    darkness_timeout: float,
    darkness_retries: int,
    darkness_url: str | None,
) -> None:
    logger.setLevel(logging.getLevelName(log_level))

    # Setup server runtime environment variables
    if darkness_url:
        os.environ["DARKNESS_TLD"] = darkness_url
    else:
        os.environ["DARKNESS_TLD"] = f"{darkness_hostname}:{darkness_port}"
    os.environ["DARKNESS_SERVICE_SLEEP_TIME"] = str(darkness_sleep_time)
    os.environ["DARKNESS_SERVICE_RETRY_COUNT"] = str(darkness_retries)
    os.environ["DARKNESS_SERVICE_TIMEOUT"] = str(darkness_timeout)
    ContextManager.deferred_init()

    logger.info("[Main] starting")
    app = FastAPI(lifespan=lifespan)

    logger.debug("[Main] adding metrics routes")
    app.include_router(metrics.router)

    logger.debug("[Main] adding Chrono routes")
    app.include_router(chrono.router)

    uvicorn.run(app, host=hostname, port=port, log_level=logging.getLevelName(log_level))


if __name__ == "__main__":
    main()



""" Entity Get Command """

import requests

from ....sdk.contracts.dtos.sdk.requests.entity.entity import EntityRequest
from ....sdk.contracts.dtos.sdk.responses.entity.get import EntityGetResponse
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ..abstract import AbstractCommand


class EntityGetCommand(AbstractCommand):
    async def execute(self, request: EntityRequest) -> EntityGetResponse:
        response: requests.Response = requests.get(
            url=f"http://{self.options.tld}/world/{request.world_id}/chunk/{request.chunk_id}/tile/{request.tile_id}/entity/{request.entity_id}",
            timeout=self.options.timeout,
        )
        return Response[EntityGetResponse].model_validate(response.json()).data



import logging

from pydantic import BaseModel
from pymongo.results import DeleteResult, UpdateResult

from ...sdk.contracts.dtos.entities.entity import Entity
from ...sdk.contracts.dtos.sdk.requests.chunk.chunk import ChunkRequest
from ...sdk.contracts.dtos.sdk.requests.chunk.create import ChunkCreateRequest
from ...sdk.contracts.dtos.sdk.requests.chunk.delete import ChunkDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.chunk.get import ChunkGetRequest
from ...sdk.contracts.dtos.sdk.requests.chunk.patch import ChunkPatchRequest
from ...sdk.contracts.dtos.sdk.requests.entity.delete import EntityDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.entity.entity import EntityRequest
from ...sdk.contracts.dtos.sdk.requests.entity.patch import EntityPatchRequest
from ...sdk.contracts.dtos.sdk.requests.tile.delete import TileDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.tile.get import TileGetRequest
from ...sdk.contracts.dtos.sdk.requests.tile.patch import TilePatchRequest
from ...sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ...sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ...sdk.contracts.dtos.sdk.requests.world.patch import WorldPatchRequest
from ...sdk.contracts.dtos.tiles.address import Address
from ...sdk.contracts.dtos.tiles.chunk import Chunk
from ...sdk.contracts.dtos.tiles.tile import Tile
from ...sdk.contracts.dtos.tiles.world import World
from ...sdk.contracts.types.connection import TileConnectionType
from ...sdk.contracts.types.dao_document import DaoDocumentType
from ..clients.dao import DaoClient
from ..factories.chunk.flat import FlatChunkFactory
from ..factories.entity.entity import EntityFactory
from ..factories.world.world import WorldFactory

logger = logging.getLogger()


# pylint: disable=too-many-public-methods
class StateService(BaseModel):
    daoclient: DaoClient

    world_factory: WorldFactory
    entity_factory: EntityFactory
    flatchunk_factory: FlatChunkFactory

    class Config:
        arbitrary_types_allowed = True

    ### Worlds ##################################

    async def worlds_get(self) -> list[World]:
        return await self.daoclient.get_all(doc_type=DaoDocumentType.WORLD)

    ### World ##################################

    async def world_create(self, request: WorldCreateRequest) -> str:
        logger.debug("[StateService] creating world")
        return await self.world_factory.create(name=request.name)

    async def world_lite_get(self, request: WorldGetRequest) -> World:
        return await self.daoclient.get(address=Address(world_id=request.id))

    async def world_get(self, request: WorldGetRequest) -> World:
        # TODO: needs conversion to daoclient backend

        # Build a complete World from Lite objects
        world: World = await self.daoclient.get(address=Address(world_id=request.id))

        partial_world = world.model_dump(exclude={"ids"})
        world: World = World.model_validate(partial_world)

        chunk_ids: set[str] = world.ids
        for chunk_id in chunk_ids:
            local_chunk: Chunk = await self.chunk_get(request=ChunkGetRequest(world_id=request.id, chunk_id=chunk_id))
            world.contents[chunk_id] = local_chunk
        return world

    async def world_delete(self, request: WorldDeleteRequest) -> bool:
        logger.debug("[StateService] deleting world")

        if request.cascade:
            world: World = await self.world_get(request=WorldGetRequest(id=request.id))
            for chunk_id in world.ids:
                await self.chunk_delete(
                    request=ChunkDeleteRequest(world_id=request.id, chunk_id=chunk_id, cascade=request.cascade)
                )

        # lastly delete the world
        result: DeleteResult = await self.daoclient.delete(address=Address(world_id=request.id))
        if result.deleted_count == 1:
            return True
        return False

    async def world_patch(self, request: WorldPatchRequest) -> UpdateResult:
        address_world: Address = Address.model_validate({"world_id": request.world_id})
        return await self.daoclient.patch(address=address_world, document=request.partial)

    ### Chunk ##################################

    async def chunk_create(self, request: ChunkCreateRequest) -> str:
        logger.debug("[StateService] creating chunk")
        new_chunk: Chunk = await self.flatchunk_factory.create(
            world_id=request.world_id, name=request.name, dimensions=request.dimensions, biome=request.biome
        )

        # Entity Factory Terrain Creation
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": new_chunk.id})
        await self.entity_factory.terrain_generate(address=address_chunk, chunk=new_chunk)

        new_chunk: Chunk = await self.daoclient.get(address=address_chunk)

        # # Entity Factory Quantum
        # await self.entity_factory.quantum(address=address_chunk)

        return new_chunk.id

    async def chunk_patch(self, request: ChunkPatchRequest) -> UpdateResult:
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": request.chunk_id})
        return await self.daoclient.patch(address=address_chunk, document=request.partial)

    async def chunk_delete(self, request: ChunkDeleteRequest) -> bool:
        msg: str = f"[WorldService] deleting chunk {id}"
        logger.debug(msg)
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": request.chunk_id})

        if request.parent:
            parent: World = await self.world_get(request=WorldGetRequest(world_id=request.world_id))
            new_ids = [id for id in parent.ids if id != request.chunk_id]
            await self.world_patch(request=WorldPatchRequest(world_id=request.world_id, partial={"ids": new_ids}))

        if request.cascade:
            chunk: Chunk = await self.chunk_get(
                request=ChunkGetRequest(world_id=request.world_id, chunk_id=request.chunk_id)
            )
            for tile_id in chunk.ids:
                await self.tile_delete(
                    request=TileDeleteRequest(
                        world_id=request.world_id,
                        chunk_id=request.chunk_id,
                        tile_id=tile_id,
                        cascade=request.cascade,
                        parent=False,
                    )
                )

        # lastly delete the chunk
        await self.daoclient.delete(address=address_chunk)

        return True

    async def chunk_lite_get(self, request: ChunkGetRequest) -> Chunk:
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": request.chunk_id})
        return await self.daoclient.get(address=address_chunk)

    async def chunk_get(self, request: ChunkGetRequest) -> Chunk:
        # Builds a complete Chunk from Lite objects
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": request.chunk_id})
        chunk: Chunk = await self.daoclient.get(address=address_chunk)

        chunk_partial = chunk.model_dump(exclude={"tile_ids"})
        chunk: Chunk = Chunk.model_validate(chunk_partial)

        # re-hydrate the tiles
        tile_ids: set[str] = chunk.ids
        for tile_id in tile_ids:
            tile: Tile = await self.tile_get(
                request=TileGetRequest(world_id=request.world_id, chunk_id=chunk.id, tile_id=tile_id)
            )

            # Add finalized tile to chunk
            chunk.contents[tile_id] = tile

        return chunk

    async def chunk_quantum(self, request: ChunkRequest):
        address: Address = Address(world_id=request.world_id, chunk_id=request.chunk_id)
        await self.flatchunk_factory.quantum(address=address)
        await self.entity_factory.quantum(address=address)

    async def chunk_quantum_tile(self, request: ChunkRequest):
        address: Address = Address(world_id=request.world_id, chunk_id=request.chunk_id)
        await self.flatchunk_factory.quantum(address=address)

    async def chunk_quantum_entity(self, request: ChunkRequest):
        address: Address = Address(world_id=request.world_id, chunk_id=request.chunk_id)
        await self.entity_factory.quantum(address=address)

    ### Tile ##################################

    async def tile_lite_get(self, request: TileGetRequest) -> Tile:
        address_tile: Address = Address.model_validate(
            {"world_id": request.world_id, "chunk_id": request.chunk_id, "tile_id": request.tile_id}
        )
        return await self.daoclient.get(address=address_tile)

    async def tile_get(self, request: TileGetRequest) -> Tile:
        tile: Tile = await self.tile_lite_get(request=request)

        # re-hydrate the entities
        entity_ids: set[str] = tile.ids
        for entity_id in entity_ids:
            entity_request: EntityRequest = EntityRequest.model_validate(
                {
                    "world_id": request.world_id,
                    "chunk_id": request.chunk_id,
                    "tile_id": request.tile_id,
                    "entity_id": entity_id,
                }
            )

            entity: Entity = await self.entity_get(request=entity_request)

            # add finalized entity to tile
            tile.contents[entity_id] = entity

        return tile

    async def tile_patch(self, request: TilePatchRequest) -> UpdateResult:
        address_tile: Address = Address.model_validate(
            {"world_id": request.world_id, "chunk_id": request.chunk_id, "tile_id": request.tile_id}
        )
        return await self.daoclient.patch(address=address_tile, document=request.partial)

    async def tile_delete(self, request: TileDeleteRequest) -> bool:
        address_tile: Address = Address.model_validate(
            {"world_id": request.world_id, "chunk_id": request.chunk_id, "tile_id": request.tile_id}
        )

        # update parent if flagged
        if request.parent:
            parent: Chunk = await self.chunk_get(
                request=ChunkGetRequest(world_id=request.world_id, chunk_id=request.chunk_id)
            )
            new_ids = [id for id in parent.ids if id != request.tile_id]
            await self.chunk_patch(
                request=ChunkPatchRequest(
                    world_id=request.world_id, chunk_id=request.chunk_id, partial={"ids": new_ids}
                )
            )

        # cascade down if flagged
        if request.cascade:
            # 1. Remove all entities
            # 2. Unhook the neighbors

            tile: Tile = await self.tile_get(
                request=TileGetRequest(world_id=request.world_id, chunk_id=request.chunk_id, tile_id=request.tile_id)
            )

            # 1. Remove all entities
            for entity_id in tile.ids:
                await self.entity_delete(
                    request=EntityDeleteRequest(
                        world_id=request.world_id,
                        chunk_id=request.chunk_id,
                        tile_id=request.tile_id,
                        entity_id=entity_id,
                        parent=False,  # Since are are removing the tile in this logic we don't need to update its parent
                    )
                )

            # 2. Unhook the neighbors
            for conn_type, neighbor_id in tile.next.items():
                neighbor: Tile = await self.tile_get(
                    request=TileGetRequest(world_id=request.world_id, chunk_id=request.chunk_id, tile_id=neighbor_id)
                )
                neigh_nexts: dict[TileConnectionType, str] = neighbor.nexts
                targ_conn_type: TileConnectionType = TileConnectionType.opposite(of=conn_type)
                if targ_conn_type not in neigh_nexts:
                    msg: str = f"conn type missing in neighbor, {conn_type} -> {targ_conn_type}"
                    logger.warning(msg)
                else:
                    del neigh_nexts[targ_conn_type]

                    # Update the neighbor
                    await self.tile_patch(
                        request=TilePatchRequest(
                            world_id=request.world_id,
                            chunk_id=request.chunk_id,
                            tile_id=neighbor_id,
                            partial={"next": neigh_nexts},
                        )
                    )

        # 3. lastly delete the tile
        await self.daoclient.delete(address=address_tile)

        return True

    ### Entity ##################################

    async def entity_get(self, request: EntityRequest) -> Entity:
        address_entity: Address = Address.model_validate(
            {
                "world_id": request.world_id,
                "chunk_id": request.chunk_id,
                "tile_id": request.tile_id,
                "entity_id": request.entity_id,
            }
        )
        return await self.daoclient.get(address=address_entity)

    async def entity_patch(self, request: EntityPatchRequest) -> UpdateResult:
        address_entity: Address = Address.model_validate(
            {
                "world_id": request.world_id,
                "chunk_id": request.chunk_id,
                "tile_id": request.tile_id,
                "entity_id": request.entity_id,
            }
        )
        return await self.daoclient.patch(address=address_entity, document=request.partial)

    async def entity_delete(self, request: EntityDeleteRequest) -> bool:
        address_entity: Address = Address.model_validate(
            {
                "world_id": request.world_id,
                "chunk_id": request.chunk_id,
                "tile_id": request.tile_id,
                "entity_id": request.entity_id,
            }
        )

        if request.parent:
            # remove self from containing tile
            tile: Tile = await self.tile_get(
                request=TileGetRequest(world_id=request.world_id, chunk_id=request.chunk_id, tile_id=request.tile_id)
            )
            new_ids: list[str] = [id for id in tile.ids if id != request.entity_id]
            await self.tile_patch(
                request=TilePatchRequest(
                    world_id=request.world_id,
                    chunk_id=request.chunk_id,
                    tile_id=request.tile_id,
                    partial={"ids": new_ids},
                )
            )

        # Lastly delete the entity
        await self.daoclient.delete(address=address_entity)

        return True



import asyncio
import logging
from asyncio import Queue

from fastapi import APIRouter
from fastapi_utils.tasks import repeat_every

from ..... import demand_env_var_as_int, get_env_var
from .....sdk.contracts.dtos.tiles.world import World
from .....sdk.contracts.types.chunk_quantum import ChunkQuantumType
from ....api.common.middleware.error import error_handler
from ..context import ContextManager

logger = logging.getLogger()

router: APIRouter = APIRouter(
    prefix="/chrono",
    tags=["chrono"],
)


@repeat_every(
    seconds=(
        demand_env_var_as_int(name="DARKNESS_WORKER_CHRONO_SLEEP_TIME")
        if get_env_var(name="DARKNESS_WORKER_CHRONO_SLEEP_TIME")
        else 10
    )
)
async def world_chrono():
    logger.debug("world chrono process executing")
    if ContextManager.client is None:
        logger.warning("Chrono client is not yet initialized")
        return

    async def producer_chunk(queue: Queue):
        worlds: list[World] = await ContextManager.client.worlds_get()
        for world in worlds:
            for chunk_id in world.ids:
                await queue.put({"world_id": world.id, "chunk_id": chunk_id})

    async def step_one():
        async def consumer(queue: Queue):
            while not queue.empty():
                address_dict: dict = await queue.get()
                await ContextManager.client.chunk_quantum(
                    world_id=address_dict["world_id"], chunk_id=address_dict["chunk_id"], scope=ChunkQuantumType.ALL
                )
                queue.task_done()

        queue = asyncio.Queue()
        await asyncio.gather(producer_chunk(queue), consumer(queue))

    await step_one()


@router.post("")
@error_handler
async def chunk_quantum() -> None:
    await world_chrono()



from fastapi import APIRouter

from .....sdk.contracts.dtos.sdk.responses.response import Response
from .....sdk.contracts.dtos.sdk.responses.worlds.get import WorldsGetResponse
from .....sdk.contracts.dtos.tiles.world import World
from ...common.middleware.error import error_handler
from ..context import ContextManager

router: APIRouter = APIRouter(
    prefix="/worlds",
    tags=["worlds"],
)


### /worlds


@router.get("")
@error_handler
async def worlds_get() -> Response[WorldsGetResponse]:
    worlds: list[World] = await ContextManager.state_service.worlds_get()
    response: WorldsGetResponse = WorldsGetResponse(worlds=worlds)
    return Response[WorldsGetResponse](data=response)
